#!/usr/bin/env python
'''
--------------------------------------------------------------------------------
    Copyright  (C) 2004, Oreste Notelli <oreste.notelli@sischema.com>          
                                                                            
    This program is free software; you can redistribute it and/or modify      
    it under the terms of the GNU General Public License as published by      
    the Free Software Foundation; either version 2 of the License, or         
    (at your option) any later version.                                       
                                                                            
    This program is distributed in the hope that it will be useful,           
    but WITHOUT ANY WARRANTY; without even the implied warranty of            
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             
    GNU Library General Public License for more details.                      
                                                                            
    You should have received a copy of the GNU General Public License     
    along with this program; if not, write to the 
    Free Software Foundation, Inc.,                                       
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             

    Sischema s.r.l.
    via Sebenico, 22
    20124 Milano - Italia
    tel-fax: (+39)0266823882
--------------------------------------------------------------------------------
    Set of classes for text parsing
    From a set of rules and a given text a python structure will be returned
    accordingly to the rules.
    For example:
    
    rules=ciclic_rule('news',
    simple_match_rule('__ignore__', 'html', 'script>'),
    simple_match_rule('__ignore__', '</html>', ''),
    fixed_rules('elem',
            [simple_match_rule ('__ignore__', '</table><a', 'href'),
             simple_match_rule ('link', '="', '"'),
             simple_match_rule ('title', '>', '</a>', html_cleaner)
            ]))

    will return the following python struct
    
    rules:
         news (list of elem)
    
    elem:
        link
        title

    so that

    for each elem in rule.news
        print elem.title
        print elem.link

'''

__version__ = "0.1"
__author__  = "Oreste Notelli <oreste.notelli@sischema.com>"

import urllib2
from string import find
from extractors import *
from utils import *

def debug (text):
    #print (text)
    pass
    
################################################################################
class base_builder:
    def __init__(self, container):
        self.container = container
    def __len__(self):
        return len(self.container)
    def result(self):
        return self.container

################################################################################

class dynamic_struct_builder(base_builder):
    def __init__(self):
        self.d= dynamic_struct()
        base_builder.__init__(self, self.d)
    def add(self, name, value):
            self.d.add(name,value)

def dynamic_struct_builder_new ():
    return dynamic_struct_builder()
        
################################################################################

def find_text_between_delimiters(body, start , end, start_pos = 0):
    ''' 
        find a text delimited by start and end
        returns a tuple startpos, endpos or None if no test is found
    '''
    res = None
    pos = find(body, start, start_pos)
    if (pos != -1):
        newpos= pos + len (start)
        endpos = find (body, end, newpos)
        if (endpos != -1):
            res = (newpos, endpos)
    return res

################################################################################

# used by rules to return field name, value and position in the text from where
# the parsing must continue
class response:
    def __init__(self, name, value, pos):
        self.name = name
        self.value = value
        self.pos = pos
    def __str__(self):
        return "name=%s\nvalue=%s\npos=%s"%(str(self.name), 
                                            str(self.value), 
                                            str(self.pos))
    def __repr__(self):
        return self.__str__()
    def __getattr__(self, name):
        if (name == self.name) : 
            return self.value
        else:
            print "Attribute :" + name
            raise AttributeError

################################################################################

class hardcoded_rule:
    ''' 
        does do anything, just return the hardcord value given in the 
        constructor 
    '''
    def __init__(self, name, value):
        self.name = name
        self.value = value
    def find (self, text, pos):
        return response(self.name, self.value, pos)


class simple_match_rule:
    ''' 
        simple rule exact match. return the text processed by the
        extractor function delimited by begin and end
        Example:
        from the text:
        
        "...e second kind of attribute references understood by instance
          objects are methods. A method is a function that ``belongs to'' 
          an object. (In Python, the term method is not unique to class 
          instances: other object ty..."

        simple_match_rule("rule", "second", "attribute")
        will return " kind of ", if you need to trim you soud pass a 
        "trim extractor" like this:

        import string
        def trim_extractor (text, bpos, epos):
            return string.strip(text[bpos:epos))
        
        and pass it to the rule
        simple_match_rule("rule", "second", "attribute", trim_extractor)
    '''
    def __init__(self, name, begin, end, extractor = simple_extractor):
        self.name = name
        self.begin = begin
        self.end = end
        self.extractor = extractor
    def find (self, text, pos):
        ''' 
        '''
        result = response (self.name, None, pos)
        res = find_text_between_delimiters(text, self.begin , self.end, pos)
        if res != None :
            result = response(self.name ,  self.extractor(text, res[0], res[1]), 
                                       res[1] + len(self.end))
        return result

################################################################################

class list_builder (base_builder):
    '''
        list builder
    '''
    def __init__(self):
        self.__l = []
        base_builder.__init__(self, self.__l)
    def add(self, name, value):
        self.__l.append(value)

def list_builder_new():
    return list_builder()

################################################################################

class dictionary_builder(base_builder):
    '''
        dictionary builder
    '''
    def __init__(self):
        self.d= {}
        base_builder.__init__(self, self.d)
    def add(self, name, value):
        self.d[name]=value

def dictionary_builder_new():
    return dictionary_builder()

################################################################################

class sequence_rules:
    ''' 
        Composed rule. Sequentially executes all the given rules
    '''
    def __init__(self, name, rules, builder_func = list_builder_new):
        self.rules = rules
        self.name = name
        self.builder_func = builder_func
    def find (self, text, pos = 0):
        result = response (self.name, None, pos)
        currentPos=pos
        builder = self.builder_func()
        for rule in self.rules:
            res = rule.find (text, currentPos)
            if (res.value != None):
                builder.add(res.name, res.value)
            currentPos = max (res.pos, currentPos)
        if (len (builder) != 0)  : result = response (name, 
                                                builder.result(),
                                                currentPos)
        return result

################################################################################

class first_in_rules:
    ''' 
        Composed rule. Sequentially executes all the given rules. Parsing 
        teminates when a rule finds a value.
    '''
    def __init__(self, name, rules):
        self.name = name
        self.rules = rules
    def find (self, text, pos):
        result = response (self.name,None, pos)
        for rule in self.rules:
            res = rule.find (text, pos)
            if (res.value != None):
                result = response (self.name, res.value, res.pos)
                break
        return result

################################################################################

class set_rules:
    '''
        Composed rule. find all given rules. If one of them is not found
        then the entire set_rel is considered not found
    '''
    def __init__(self, name, rules, builder_func= dynamic_struct_builder_new):
        self.name = name
        self.rules = rules
        self.builder_func = builder_func
    def find (self, text, pos):
        result = response(self.name, None, pos)
        currentPos=pos
        found = False
        builder = self.builder_func()
        for rule in self.rules:
            res = rule.find (text, currentPos)
            if (res.value == None):
                found = False
                break
            builder.add (res.name,res.value)
            currentPos = max (res.pos, currentPos)
            found = True
        if (found): result = response(self.name, builder.result(), 
                                      currentPos)
        return result


################################################################################

class cyclic_rule:
    '''
        Iterates from the starting rule, until the end rule text position 
        or if more cyclicRule founds values.
    '''
    def __init__(self, name, startRule, endRule, cyclicRule, 
                 builder_func = list_builder_new):
        self.name = name
        self.startRule = startRule
        self.endRule = endRule
        self.cyclicRule = cyclicRule
        self.builder_func = builder_func
    def find (self, text, pos):
        result = response(self.name, None, pos)
        currentPos = pos
        startRes = self.startRule.find(text, pos)
        builder = self.builder_func()
        debug ("ciclic_rule.find startRes= " + str(startRes))
        if (startRes.value!= None):
            startpos = startRes.pos
            endRes = self.endRule.find(text, startpos)
            debug ("ciclic_rule.find endRes= " + str(endRes))
            if (endRes.value != None) :
                endPos = endRes.pos
                currentPos = startpos
                while currentPos < endPos:
                    debug ("ciclic_rule.find currentPos= " + str(currentPos))
                    cyclRes = self.cyclicRule.find(text, currentPos)
                    debug ("ciclic_rule.find cyclRes= " + str(cyclRes))
                    currentPos = max (currentPos, cyclRes.pos)
                    if (cyclRes.value == None):  break
                    builder.add(cyclRes.name, cyclRes.value)
        if ((len(builder)) != 0): result= response (self.name, 
                                                         builder.result(),
                                                         currentPos)
        return result


################################################################################

def process_rule(rule, text):
    ''' 
        process the given rule on the given text.
        returns the rule defined value.
    '''
    parsed_result = rule.find(text, 0)
#    print ("process_rule: " + str(type(parsed_result)) + " " + str(parsed_result))
    return parsed_result.value
        




        
